

// This program is an example of a GDK application.	It accepts requests
// from clients and then passes those requests on to an SQL server.  The
// results from the SQL server are then sent back to the client.

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>

#define INCL_BASE
#include	<OS2.h>

#include	<sqlfront.h>
#include	<sqldb.h>

#include	<srv.h>

// function prototypes
//
void main( int argc, char **argv );

int	SRVAPI init_server	( SRV_SERVER FAR * server );

void	SRVAPI set_remote_server_name ( char FAR * name );

int	chk_err      ( SRV_SERVER FAR * server,
		       SRV_PROC   FAR * srvproc,
		       int		srverror,
		       BYTE		severity,
		       BYTE		state,
		       int		oserrnum,
		       DBCHAR	  FAR * errtext,
		       int		errtextlen,
		       DBCHAR	  FAR * oserrtext,
		       int		oserrtextlen );

DBCHAR	FAR * getservername( int argc, char * argv[] );

void main(argc, argv)
int	argc;
char	*argv[];
{
    SRV_CONFIG	FAR *config;	    // The GDK configuration structure
    SRV_SERVER	FAR *server;	    // The GDK service process
    DBCHAR	FAR *GDKsrvrname;   // The server name of this GDK

    //	Read any command line arguments.
    //
    GDKsrvrname = getservername(argc, argv);

    //	Send the name retrieved to the gateway's DLL module
    //
    set_remote_server_name( GDKsrvrname );

    //	Allocate a configuration structure that is used to initialize
    //	the GDK application
    //
    config = srv_config_alloc();

    //	Allow 20 connections at a time.
    //
    srv_config( config, (DBINT)SRV_CONNECTIONS, "20", SRV_NULLTERM );

    //	Set the log file.
    //
    srv_config( config, (DBINT)SRV_LOGFILE, "gateway.log", SRV_NULLTERM );

    //	Install the GDK error handler.
    //
    srv_errhandle( chk_err );

    //	Initialize the gateway GDK and save the server handle
    //	so it can be used in later functions.
    //
    server = srv_init( config, TDS34_CONNECT_NAME, SRV_NULLTERM );

    if( server == NULL )
    {
	printf( "Unable to initialize GDK.  Check log file.\n" );
	DosExit( EXIT_PROCESS, 1 );
    }

    // When starting the gateway, initialize the remote server structure.
    // This is done in the init_server() function.
    // All the other event handlers are also defined in the init_server()
    // function.
    //
    srv_handle(server, (DBINT)SRV_START, init_server);

    //	Now everything's ready to go with our GDK, so we
    //	start it and keep it going until we get a stop request.
    //
    srv_log( server, FALSE, " ", SRV_NULLTERM );    // insert blank line
    srv_log( server, TRUE,  "GDK Starting", SRV_NULLTERM );

    printf( "\nGDK Starting\n" );

    srv_run(server);

}


//  GETSERVERNAME - getservername( argc, argv )
//
//  Read the server name from the command the line.  If no name is supplied,
//   then a local named pipe is assumed.
//
//  Inputs:
//	argc - int (from "main" entry)
//	argv - pointer to array of char pointers (from "main" entry)
//
//  Outputs:
//	Returns a pointer to the server name supplied on the command line.
//
DBCHAR FAR * getservername(argc, argv)
int	argc;
char	*argv[];
{
    if (argc > 1 )
	return( argv[1] );
    else
	return( NULL );
}
