// This module defines all of the "callback" functions.  These modules
// are defined here due to the fact that they must be mult-threaded and
// therefore completely reentrant.  They must be linked with the proper
// C Runtime Multi-Threaded Library (LLIBCDLL.LIB).

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#define INCL_BASE
#include <OS2.h>

#include <sqlfront.h>
#include <sqldb.h>

#include <srv.h>

#define INCL_NETMAILSLOT
#define INCL_NETSTATS
#define INCL_NETERRORS
#include <lan.h>

//  Define some GDK user message codes.
//
#define  SRV_MAXERROR   20000
#define  RPC_UNKNOWN   SRV_MAXERROR + 1
#define  INVALID_RPC_SYNTAX   SRV_MAXERROR + 2
#define  BAD_RPC_PARAMETER   SRV_MAXERROR + 3
#define  BROADCAST SRV_MAXERROR + 4
#define  EXEC SRV_MAXERROR + 5

//  Miscellaneous defines used by rpc-handling routines.
//
#define  EXEC_CMD "exec"
#define  MAXLEN   80
#define  MAXARGS  4


// Misc defines for simple data file stored procedure
#define BUF_SIZE	400
#define MAX_NAME_LENGTH	60
#define MAX_ADDR_LENGTH 80



// Event handlers for RPC requests.
//
RETCODE rpc_list();            // List RPC associated with GDK Server
RETCODE rpc_broadcast();       // Broadcast message to specifed mailslot
RETCODE rpc_exec();            // Execute the specified command string
RETCODE rpc_list();            // List RPC associated with GDK Server
RETCODE rpc_disk();            // Return disk drives and their available space
RETCODE rpc_test();
int	GetStatistics ( SRV_PROC FAR * srvproc );

int	GetStatistics2 ( SRV_PROC FAR * srvproc );
RETCODE GetCustomerInfo();	   // stored procedure for reading simple file format

// file name for file stored procedure
char	*pchFileName = "dba_test.db";





// RPC argument information structure.
//
typedef struct rpc_args
{
   DBCHAR      name[MAXNAME+1];
   int         type;
   DBINT       length;
   DBCHAR      defaultvalue[MAXLEN];
} RPCARGS;

// RPC information structure.
//
typedef struct rpc
{
   DBCHAR      name[MAXNAME+1];
   DBCHAR      usage[MAXLEN];
   int         numparams;
   RPCARGS     params[MAXARGS];
   RETCODE     (FAR * handler) (void FAR *, ...);
} RPC;

// Array of RPC handlers.
//
RPC Rpcs[] =
{
   "rpc_exec",                                  // rpc name
      "usage: rpc_exec <command> <parameters>", // rpc usage
       2,                                       // number of parameters
         {                                      // parameter definitions

            "command",                          // parameter name
               SRVCHAR,                         // parameter type
               MAXLEN,                          // parameter length (0 if fixed)
               "bldmastr.exe",                  // default parameter value

            "parameters",                       // parameter name
               SRVCHAR,                         // parameter type
               MAXLEN,                          // parameter length (0 if fixed)
               "/d c:\default.dat /s 7158",     // default parameter value
         },
      rpc_exec,                                 // rpc function pointer

   "rpc_list",
      "usage: rpc_list",
      0,
         {0},
      rpc_list,

   "rpc_disk",
      "usage: rpc_disk",
      0,
         {0},
      rpc_disk,

   "rpc_broadcast",
      "usage: rpc_broadcast <mailslot> <message>",
      2, {
            "mailslot",
               SRVCHAR,
               MAXLEN,
               "",
            "message",
               SRVCHAR,
               MAXLEN,
               "",
         },
      rpc_broadcast,

   "rpc_test",                            
      "usage: rpc_test <value>",          
       1,                                 
         {                                
            "arg1",                       
               SRVFLT8,                   
               0,                         
               "0.666",                   
         },
      rpc_test,

	"sp_lan_statistics",
      "usage: sp_lan_statistics",
      0,
         {0},
      GetStatistics,

	"sp_lan_statistics2",
      "usage: sp_lan_statistics2",
      0,
         {0},
      GetStatistics2,

	"get_customer_info",
      "usage: get_customer_info",
      0,
         {0},
      GetCustomerInfo,






};
#define Rpcnumber sizeof(Rpcs) / sizeof(RPC)

#define STAT_ARRAY_SIZE 100
// define a structure for the LANMAN statistics
typedef struct _LAN_STATS {
char    szServerName[CNLEN+1];
time_t  Timestamp;
struct stat_server_0	StatServ0;
}   LAN_STATS;

LAN_STATS   aLanStats[STAT_ARRAY_SIZE];

HSEM   				StatsSem;
int    				giCurrentRow = 0;
unsigned long		gulSleepTime = 60000;
BYTE   				ThreadStack[8000];

//  Prototype dblib API's
//
void dbsetuserdata(DBPROCESS FAR * dbproc, void FAR * data);

void FAR * dbgetuserdata(DBPROCESS FAR * dbproc);

RETCODE dbcolntype (DBPROCESS FAR * dbproc, int column);

RETCODE attn_handler (SRV_PROC FAR * srvproc);

RETCODE chk_err (SRV_SERVER FAR * server,
              SRV_PROC FAR * srvproc,
              int srverror,
              BYTE severity,
              BYTE state,
              int oserrnum,
              DBCHAR FAR * errtext,
              int errtextlen,
              DBCHAR FAR * oserrtext,
              int oserrtextlen);

RETCODE init_remote (SRV_PROC FAR * srvproc);

RETCODE init_server (SRV_SERVER FAR * server);

RETCODE rpc_execute (SRV_PROC FAR * srvproc);

RETCODE exit_remote (SRV_PROC FAR * srvproc);

RETCODE lang_execute (SRV_PROC FAR * srvproc);

void set_remote_server_name(char FAR * name);

void	StatDaemon();


//  Miscellaneous prototypes for rpc language events
//
RETCODE lang_parser(SRV_PROC FAR * srvproc,
                     RPC         ** rpc,
                     DBCHAR         paramvalues[MAXARGS][MAXLEN]);

DBCHAR * scan_next(DBCHAR * string,
                    DBCHAR * word);

//
//  This section defines the "call-back" functions supplied to the gateway
//  application.
//
#pragma check_stack(off)  // turn off stack checking


//
//  SET_REMOTE_SERVER_NAME
//
//   This function sets the name of the remote server.
//
//  Parameters:
//   Pointer to name of server.
//
//  Returns:
//   none
//
void set_remote_server_name(name)
char FAR *name;
{
   return;
}

//
//  INIT_SERVER
//   Initialize the server on a SRV_START event.
//   Event handlers for the server are installed.
//
//  Parameters:
//   Pointer to SRV_SERVER structure
//
//  Returns:
//   SRV_CONTINUE
//
RETCODE init_server(server)
SRV_SERVER FAR *server;
{
    char log_buffer[256];
    PVOID   pStatDaemonStack = ThreadStack + 8000;
    TID     StatTID;


    //   When we get a connection request from a client, we want to
    //   call "init_remote()" to make a connection to the remote
    //   server.
    //
    srv_handle(server, (DBINT)SRV_CONNECT, init_remote);

    // When the client issues a language request, call
    // "lang_execute()" to send the SQL statement to the remote DBMS.
    //
    srv_handle(server, (DBINT)SRV_LANGUAGE, lang_execute);

    // When the client issues an RSP, call "rpc_execute()"
    // to send the RSP to the remote DBMS (the SQL Server).
    //
    srv_handle(server, (DBINT)SRV_RPC, rpc_execute);

    // When a disconnect request is issued, call "exit_remote()"
    // to close the connection to the remote DBMS.
    //
    srv_handle(server, (DBINT)SRV_DISCONNECT, exit_remote);

    //   Log Server information to log file
    //
    sprintf(log_buffer,
      "Server pipe name = %s",
      srv_sfield(server, SRV_SERVERNAME, (int *)NULL));

    srv_log(server, FALSE, log_buffer, SRV_NULLTERM);
    printf( "%s\n", log_buffer);

    sprintf(log_buffer,
        "Client connections allow = %s",
        srv_sfield(server, SRV_CONNECTIONS, (int *)NULL));

    srv_log(server, FALSE, log_buffer, SRV_NULLTERM);
    printf( "%s\n", log_buffer);

	// now start up the LAN statistics data collection thread

    DosCreateThread(StatDaemon, &StatTID, pStatDaemonStack);


    return (SRV_CONTINUE);
}


//  INIT_REMOTE
//  Event handler for a SRV_CONNECT event.
//  A connection is made to the remote DBMS.
//
//  In this example program, the connection is to a SQL Server.
//  If using a non-SQL Server remote DBMS, the connection code
//  would be different, but would probably still occur here.
//
//  Parameters:
//   srvproc - the handle to the client connection that got the SRV_CONNECT.
//
//  Returns:
//   SRV_CONTINUE
//
//  Side Effects:
//   If the connection to the remote dbms cannot be made, then issue
//   a SRV_DISCONNECT request.
//
//
RETCODE init_remote(srvproc)
SRV_PROC FAR *srvproc;
{

    //   Display connect info on console
    //
    printf( "\nClient process ID: %s\n",
        srv_pfield(srvproc, SRV_CPID, (int *) NULL));

    printf( "User name: %s\n",
        srv_pfield(srvproc, SRV_USER, (int *) NULL));

    printf( "Application program name: %s\n",
        srv_pfield(srvproc, SRV_APPLNAME, (int *) NULL));

    return (SRV_CONTINUE);
}


//
//  LANG_EXECUTE
//   Execute a client language request on the remote DBMS.
//   Any results from the remote DBMS are passed back to the GDK client.
//
//  Parameters:
//   srvproc - GDK process handle to the current client connection.
//
//  Returns:
//   SRV_CONTINUE
//

RETCODE lang_execute(srvproc)
SRV_PROC FAR *srvproc;
{
   int      i;
   RETCODE  status;                          // status of dblib calls
   DBCHAR   paramvalues[MAXARGS][MAXLEN];
   BYTE     convertvalues[MAXARGS][MAXLEN];
   RPC      *rpc = NULL;

   /* Initialize parameter storage */
   for (i= 0; i< MAXARGS; i++){
      memset(paramvalues[i], 0, MAXLEN);
      memset(convertvalues[i], 0, MAXLEN);
   }

   if (lang_parser(srvproc, &rpc, paramvalues) == SUCCEED)
   {
      for (i = 0; i < rpc->numparams; i++)
      {
         if (strlen(paramvalues[i]) == 0 && strlen(rpc->params[i].defaultvalue) == 0)
         {
            srv_sendstatus(srvproc, 1);
            srv_sendmsg(srvproc, SRV_MSG_ERROR,
                        INVALID_RPC_SYNTAX,
                        SRV_INFO,
                        (DBTINYINT)0,
                        NULL,
                        0,
                        0,
                        rpc->usage,
                        SRV_NULLTERM);
            srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_FINAL), 0, 0);
            return (SRV_CONTINUE);
         }

         if (strlen(paramvalues[i]) == 0 && strlen(rpc->params[i].defaultvalue) != 0)
            strcpy (paramvalues[i], rpc->params[i].defaultvalue);

         // convert parameters from character string to parmeter type
         //
         srv_convert(srvproc,
                     SRVCHAR,
                     paramvalues[i], 
                     -1,
                     rpc->params[i].type,
                     &convertvalues[i],
                     rpc->params[i].length);
      }

      status = (rpc->handler)(srvproc,
                              &convertvalues[0],
                              &convertvalues[1],
                              &convertvalues[2],
                              &convertvalues[3]);
   }
   return (SRV_CONTINUE);
}

RETCODE lang_parser(srvproc, rpc, paramvalues)
SRV_PROC FAR *srvproc;
RPC          **rpc;
DBCHAR       paramvalues[MAXARGS][MAXLEN];
{
   DBCHAR FAR     *query;       // pointer to language buffer
   int            i;
   int            numparams;
   DBCHAR         msg[MAXLEN * 2];
   DBINT          msgnum;
   DBCHAR         rpcname[MAXLEN];
   BOOL           paramsyntax = FALSE;
   DBCHAR         paramname[MAXLEN];
   DBCHAR         equalstring[2];
   DBCHAR         *paramvalue = NULL;

   query = srv_langptr(srvproc);

   query = scan_next(query, rpcname);
   if (strlen(rpcname) == 0)
   {
      srv_senddone(srvproc, SRV_DONE_FINAL, 0, 0);
      return (FAIL);
   }

   if(strnicmp(rpcname, EXEC_CMD, (sizeof(EXEC_CMD) - 1)) == 0)
   {
      // rpc name 
      //
      query = scan_next(query, rpcname);
      if (strlen(rpcname) == 0)
         goto syntax_error;
   }

   // Check for existence 
   //
   for (i = 0; i < Rpcnumber; i++)
      if (strcmp(Rpcs[i].name, rpcname) == 0)
      {
         *rpc = &Rpcs[i];
         break;
      }
   if (*rpc == NULL)
   {
      sprintf(msg, "Procedure \'%s \' not found.", rpcname);
      msgnum = RPC_UNKNOWN;
      goto error;
   }

   // parameters
   //
   numparams = 0;
   while (*query != '\0')
   {
      if (++numparams > (*rpc)->numparams)
      {
         sprintf(msg, (*rpc)->usage);
         msgnum = INVALID_RPC_SYNTAX;
         goto error;
      }

      if (!paramsyntax && *query == '@')     
         paramsyntax = TRUE;                 // parameter name mode

      if (!paramsyntax)
         if (paramvalue == NULL)
            paramvalue = paramvalues[0];
         else
            paramvalue += MAXLEN;

      if (paramsyntax)
      {
         if (*query != '@')
         {
            sprintf(msg, "Once the form '@name = value' has been used, all subsequent parameters must be passed in the form '@name = value'.");
            msgnum = INVALID_RPC_SYNTAX;
            goto error;
         }
         else
            query++;

         query = scan_next(query, paramname);
         if (strlen(paramname) == 0)
            goto syntax_error;

         // Get parameter index 
         //
         paramvalue = NULL;
         for (i = 0; i < (*rpc)->numparams; i++)
            if(strcmp((*rpc)->params[i].name, paramname) == 0)
            {
               paramvalue = paramvalues[i];
               break;
            }
         if (paramvalue == NULL)
         {
            sprintf(msg,
                    "Procedure '%s' does not recognize parameter name: %s",
                    rpcname,
                    paramname);
            msgnum = BAD_RPC_PARAMETER;
            goto error;
         }

         // Already assigned value
         //
         if(strlen(paramvalue) > 0)
            goto syntax_error;

         // Check for '='
         //
         query = scan_next(query, equalstring);
         if (*equalstring != '=')
            goto syntax_error;

      }

      query = scan_next(query, paramvalue);
      if (strlen(paramvalue) == 0)
         goto syntax_error;

      if (*query == ',')
      {
         query++;   
         while (*query == ' ' || *query == '\t') query++;
      }
   }
   return (SUCCEED);

syntax_error:
      sprintf(msg, "Incorrect syntax found near '%s'.", query);
      msgnum = INVALID_RPC_SYNTAX;

error:
   srv_sendstatus(srvproc, 1);
   srv_sendmsg(srvproc, SRV_MSG_ERROR,
               msgnum,
               SRV_INFO,
               (DBTINYINT)0,
               NULL,
               0,
               0,
               msg,
               SRV_NULLTERM);
   srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_FINAL), 0, 0);
   return (FAIL);
}

DBCHAR * scan_next(string, word)
DBCHAR   *string;
DBCHAR   *word;
{
   DBCHAR  *p;

   word[0] = '\0';
   if (*string == '\"' || *string == '\'')
   {
      // check for unclosed quote 
      //
      p = strchr(string + 1, *string);
      if (p == NULL)
         return(string);
      strncpy(word, string + 1, p - (string + 1));
      word[p - (string + 1)] = '\0';
      string += 2; 
   }
   else
   {
      // clear proceeding white space
      // 
      while (*string == ' ' || *string == '\t' || *string == '\n') string++;
      sscanf(string, "%s", word);

      // ignore comments
      //
      while (strncmp(word, "/*", 2) == 0)
      {
         string = strstr(string, "*/");
         if (string != NULL){
            string += 2;
            word[0] = '\0';
            while (*string == ' ' || *string == '\t' || *string == '\n') string++;
            sscanf(string, "%s", word);
         }
         else
            return(string);
      }
   }
 
   if (strlen(word) > 0)
      string += strlen(word);

   // clear trailing white space
   // 
   while (*string == ' ' || *string == '\t' || *string == '\n') string++;
   
   return(string);
}

   
//
//  RPC_EXECUTE
//   Execute a client RSP.
//
//   Note that in the example, the remote DBMS is an SQL Server.
//   If the remote DBMS was a different DBMS, the code to send
//   the query to the remote DBMS and receive remote results
//   would be different.  In this example, RSP is sent
//   along to the SQL Server.
//
//  Parameters:
//   srvproc - The GDK process handle to use to send results to the client.
//
//  Returns:
//   SRV_CONTINUE
//
RETCODE rpc_execute(srvproc)
SRV_PROC FAR *srvproc;
{
   int     i;
   int     x;
   int     y;
   int     len;
   int     numparams;
   RETCODE status;
   RPC     *rpc;
   DBCHAR  msg[MAXLEN];
   DBINT   msgnum;
   DBCHAR  paramvalues[MAXARGS][MAXLEN];
   BOOL    paramnamemode = FALSE;
   DBCHAR  *paramname;
   DBCHAR  *value;
   DBINT   type;

   // Initialize parameter storage 
   //
   for (i= 0; i< MAXARGS; i++){
      memset(paramvalues[i], 0, MAXLEN);
   }

   for (i = 0; i < Rpcnumber; i++)
      if (strcmp(Rpcs[i].name, srv_rpcname(srvproc, (int *) NULL))
         == 0)
      {
         rpc = &Rpcs[i];
         numparams = srv_rpcparams(srvproc);
         if (srv_paramname(srvproc, 1, &len) && len > 0)
            paramnamemode = TRUE;
         for (y = 1; y <= numparams; y++)
         {
            /* Find parameter number */
            if (paramnamemode)
            {
               paramname = srv_paramname(srvproc, y, &len);
               if (strlen(paramname) == 0)
                  goto parameter_error;

               if (*paramname == '@')
                  paramname++;
               else
                  goto parameter_error;
               value = NULL;
               for (x = 0; x < rpc->numparams; x++)
                  if(strcmp(rpc->params[x].name, paramname) == 0)
                  {
                     value = paramvalues[x];
                     type = rpc->params[x].type;
                     break;
                  }
               if (value == NULL)
                  goto parameter_error;
            }
            else
            {
               value = paramvalues[y - 1];
               type = rpc->params[y - 1].type;
            }

            // Check parameters for correct type
            //
            if (type != srv_paramtype(srvproc, y))
            {
               sprintf(msg, "Parameter \'%s \' is incorrect type.", 
                       paramname);
               msgnum =  BAD_RPC_PARAMETER;
               goto error;
            }

            // Move RPC parameters to local variables
            //
            srv_bmove(srv_paramdata(srvproc, y), value, srv_paramlen(srvproc, y));
            value[srv_paramlen(srvproc, y)] = '\0';

          }
      }

   // If unspecified, use default value 
   //
   for (i = 0; i < rpc->numparams; i++)
   {
      if (strlen(paramvalues[i]) == 0 && strlen(rpc->params[i].defaultvalue) == 0)
      {
         strcpy(msg,rpc->usage);
         msgnum = INVALID_RPC_SYNTAX;
         goto error;
      }

      if (strlen(paramvalues[i]) == 0 && strlen(rpc->params[i].defaultvalue) != 0)
         strcpy (paramvalues[i], rpc->params[i].defaultvalue);
   }

   status = (*rpc->handler)(srvproc,
                            paramvalues[0],
                            paramvalues[1],
                            paramvalues[2],
                            paramvalues[3]);
   return (SRV_CONTINUE);

   sprintf(msg, "Procedure \'%s \' not found.", 
           srv_rpcname(srvproc, (int *) NULL));
   msgnum = RPC_UNKNOWN;
   goto error;

parameter_error:
   sprintf(msg, "Procedure '%s' does not recognize parameter name: %s",
           rpc->name,
           paramname);
   msgnum =  BAD_RPC_PARAMETER;

error:
   srv_sendstatus(srvproc, 1);
   srv_sendmsg(srvproc, SRV_MSG_ERROR,
               msgnum,
               SRV_INFO,
               (DBTINYINT)0,
               NULL,
               0,
               0,
               msg,
               SRV_NULLTERM);
   srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_FINAL), 0, 0);
   return (SRV_CONTINUE);
}


//  EXIT_REMOTE
//   Handler for SRV_DISCONNECT events.
//   Closes remote DBMS connection if appropriate.
//
//  The code to disconnect from the remote DBMS is SQL Server specific. If
//  using a non-SQL Server database, the disconnect from the  remote database
//  would be different but would probably still occur here.
//
//  Parameters:
//   srvproc - the handle to the client connection
//
//  Returns:
//   SRV_DISCONNECT
///
RETCODE exit_remote(srvproc)
SRV_PROC FAR *srvproc;
{
    //   Display info on console
    //
    printf( "\nClient connection closed, process ID: %s\n",
      srv_pfield(srvproc, SRV_CPID, (int *) NULL));

    return (SRV_CONTINUE);
}


//  CHK_ERR
//   Print out GDK errors.
//
//  Parameters:
//   server        - pointer to gateway server structure.
//   srvproc      - pointer to client connection structure
//   errornum     - error number.
//   severity     - error severity.
//   state        - error state.
//   oserrnum     - operating system error number, if any.
//   errtext      - the text of the error message.
//   errtextlen   - length of the errtext message
//   oserrtext    - the text of the operating system error message.
//   oserrtextlen - length of the errtext message
//
//  Returns:
//   SRV_CONTINUE, SRV_CANCEL, or SRV_EXIT_PROGRAM
//

RETCODE chk_err(server, srvproc, errornum, severity, state, oserrnum, errtext,
            errtextlen, oserrtext, oserrtextlen)
SRV_SERVER FAR *server;
SRV_PROC FAR   *srvproc;
int            errornum;
BYTE           severity;
BYTE           state;
int            oserrnum;
DBCHAR FAR     *errtext;
int            errtextlen;
DBCHAR FAR     *oserrtext;
int            oserrtextlen;
{
    char    log_buffer[256];

    //   Operating system error?
    //
    if (oserrnum != SRV_ENO_OS_ERR)
    {
      sprintf(log_buffer,
         "SERVER OS ERROR: %d: %s.",
         oserrnum,
         oserrtext);

      srv_log(server, TRUE, log_buffer, SRV_NULLTERM);
    }

    //   Is this a fatal error for the GDK server?
    //
    if (severity >= SRV_FATAL_SERVER)
    {
      sprintf(log_buffer,
         "SERVER: FATAL SERVER ERROR: errornum = %d, "
         "severity = %d, state = %d: %s.",
         errornum,
         severity,
         state,
         errtext);

      srv_log(server, TRUE, log_buffer, SRV_NULLTERM);
      return (SRV_EXIT);
    }
    else {
      //
      //  Did the "srvproc" get a fatal error?
      //
      if (severity >= SRV_FATAL_PROCESS)
      {
         sprintf(log_buffer,
            "SERVER: FATAL CONNECT ERROR: errornum = %d, "
            "severity = %d, state = %d: %s.",
            errornum,
            severity,
            state,
            errtext);

         srv_log(server, TRUE, log_buffer, SRV_NULLTERM);

         //
         //   Make a disconnect request, to be executed if possible.
         //
         // if (srvproc != (SRV_PROC *) NULL)
         //          srv_event(srvproc, SRV_DISCONNECT, (BYTE *) NULL);

         return (SRV_CANCEL);
      }
    }

    //   A non-fatal error or an information message received.
    //   We'll pass it through to the client.
    //
    if (srvproc != (SRV_PROC *) NULL && (server != NULL))

      if(severity < 10) {          // if informational message
         srv_sendmsg(srvproc,
            SRV_MSG_INFO,
            (DBINT)errornum,
            severity,
            0,
            NULL,
            0,
            0,
            errtext,
            SRV_NULLTERM);
      }
      else {                // must be an error message
         srv_sendmsg(srvproc,
            SRV_MSG_ERROR,
            (DBINT)errornum,
            severity,
            0,
            NULL,
            0,
            0,
            errtext,
            SRV_NULLTERM);
      }
      else {
         sprintf(log_buffer,
            "GATEWAY ERROR: errornum = %d, severity = %d: %s",
            errornum,
            severity,
            errtext);

         srv_log(server, TRUE, log_buffer, SRV_NULLTERM);
      }

   return (SRV_CONTINUE);
}

RETCODE rpc_list(srvproc)
SRV_PROC FAR *srvproc;
{
   DBCHAR colname1[MAXNAME+1]; 
   DBCHAR colname2[MAXNAME+1]; 
   int  i;

   sprintf(colname1, "rpcname");
   srv_describe(srvproc, 1, colname1, SRV_NULLTERM, SRVCHAR, MAXNAME+1,
      SRVCHAR, 0, NULL);

   sprintf(colname2, "rpcusage");
   srv_describe(srvproc, 2, colname2, SRV_NULLTERM, SRVCHAR, MAXLEN,
      SRVCHAR, 0, NULL);

   // Return each RPC handler as a row
   //
   for (i = 0; i < Rpcnumber; i++)
   {
      srv_setcoldata(srvproc, 1, Rpcs[i].name);
      srv_setcollen(srvproc, 1, strlen(Rpcs[i].name));

      srv_setcoldata(srvproc, 2, Rpcs[i].usage);
      srv_setcollen(srvproc, 2, strlen(Rpcs[i].usage));

      srv_sendrow(srvproc);
   }
   srv_senddone(srvproc, (SRV_DONE_COUNT | SRV_DONE_FINAL), 0, Rpcnumber);
   
   return(SUCCEED);
}


RETCODE rpc_broadcast(srvproc, mailslot, msg)
SRV_PROC FAR *srvproc;
DBCHAR *mailslot;
DBCHAR *msg;
{
   unsigned retcode;

   // Send message as second class mail
   retcode = DosWriteMailslot(mailslot, msg, strlen(msg), 0, 2, 0);
   if (retcode)
   {
      srv_sendstatus(srvproc, 1);
      srv_sendmsg(srvproc, SRV_MSG_ERROR,
                  BROADCAST,
                  SRV_INFO,
                  (DBTINYINT)0,
                  NULL,
                  0,
                  0,
                  "Unable to broadcast message due to error.",
                  SRV_NULLTERM);
      srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_FINAL), 0, 0);
      return(FAIL);
   }

   // Broadcast successful
   //
   srv_sendstatus(srvproc, 0);
   srv_sendmsg(srvproc,
               SRV_MSG_INFO,
               BROADCAST,
               (DBTINYINT)0,
               (DBTINYINT)0,
               NULL,
               0,
               0,
               "Broadcast complete.",
               SRV_NULLTERM);
    srv_senddone(srvproc, (SRV_DONE_COUNT | SRV_DONE_FINAL), 0, 0);
    return(SUCCEED);
}

RETCODE rpc_exec(srvproc, command, params)
SRV_PROC FAR *srvproc;
DBCHAR *command;
DBCHAR *params;
{
   DBCHAR      comstr[MAXLEN];
   DBCHAR      modulename[MAXLEN];
   DBCHAR      msg[MAXLEN * 2];
   int         len;
   RESULTCODES retcode;

   strcpy(comstr, command);
   len = strlen(comstr);
   comstr[len] = '\0';
   strcpy(&comstr[len + 1], params);
   comstr[strlen(&comstr[len + 1]) + len + 1] = '\0';
   comstr[strlen(&comstr[len + 1]) + len + 2] = '\0';

   sprintf(msg, "Executing '%s %s.'", command, params),
   srv_sendmsg(srvproc,
               SRV_MSG_INFO,
               EXEC,
               (DBTINYINT)0,
               (DBTINYINT)0,
               NULL,
               0,
               0,
               msg,
               SRV_NULLTERM);
   srv_senddone(srvproc, SRV_DONE_MORE, 0, 0);

   // Execute string
   //
   if (DosExecPgm(modulename, 
               sizeof(modulename), 
               EXEC_SYNC, 
               comstr, 
               NULL, 
               &retcode,
               comstr) != 0)
   {
      srv_sendstatus(srvproc, 1);
      srv_sendmsg(srvproc,
                  SRV_MSG_ERROR,
                  EXEC,
                  SRV_INFO,
                  (DBTINYINT)0,
                  NULL,
                  0,
                  0,
                  "Error executing command",
                  SRV_NULLTERM);
      srv_senddone(srvproc, (SRV_DONE_ERROR | SRV_DONE_FINAL), 0, 0);
      return(FAIL);
   }

   // Command successful
   //
   srv_sendstatus(srvproc, 0);
   srv_sendmsg(srvproc,
               SRV_MSG_INFO,
               EXEC,
               (DBTINYINT)0,
               (DBTINYINT)0,
               NULL,
               0,
               0,
               "Command complete.",
               SRV_NULLTERM);
    srv_senddone(srvproc, (SRV_DONE_COUNT | SRV_DONE_FINAL), 0, 0);
    return(SUCCEED);
}

RETCODE rpc_disk(srvproc)
SRV_PROC FAR *srvproc;
{
   DBCHAR           colname1[MAXNAME+1]; 
   DBCHAR           colname2[MAXNAME+1]; 
   DBCHAR           drivename;
   int            drivenum;
   unsigned short disk;
   unsigned long  drivenums;
   FSALLOCATE     data;
   long           space_remaining;
   int            i = 0;

   sprintf(colname1, "drive");
   srv_describe(srvproc, 1, colname1, SRV_NULLTERM, SRVCHAR, 1, 
                SRVCHAR, 1, (BYTE *) &drivename);

   sprintf(colname2, "bytes free");
   srv_describe(srvproc, 2, colname2, SRV_NULLTERM, SRVINT4, 4, 
      SRVINT4, 4, (BYTE *) &space_remaining);

   DosQCurDisk(&disk, &drivenums);

   drivenums >>= 2;   //Ignore drives A and B
   for (drivename = 'C', drivenum = 3; drivename <= 'Z'; drivename++, drivenum++)
   {
      if (drivenums & 1)
      {
         i++;
         DosQFSInfo(drivenum,1,(char far *) &data, sizeof(data));
         space_remaining = (long)data.cSectorUnit * (long)data.cUnitAvail;
         space_remaining *= (long) data.cbSector;
         srv_sendrow(srvproc);
      }
      drivenums >>= 1;
   }
   srv_senddone(srvproc, (SRV_DONE_COUNT | SRV_DONE_FINAL), 0, i);
   return(SUCCEED);
}

//test
RETCODE rpc_test(srvproc, test)
SRV_PROC *srvproc;
DBFLT8  *test;
{
   printf("TEST: %.2f", *test);
   srv_senddone(srvproc, SRV_DONE_FINAL, 0, 0);
   return(SUCCEED);
}

	

/*
** Name:  GetStatistics()
**
** Purpose:  
**
**
*/
                                 
int	GetStatistics2(pSrvProc)
SRV_PROC    FAR *pSrvProc;
{	
    char            *pszService = "SERVER";
    char            *pbBuffer;      // pointer to data buffer
    API_RET_TYPE    uStatus;
    int             i, iCount;         // Index for arguments
    unsigned short  cbBuffer;       // size of buffer
    unsigned short  cbTotalAvail;   // count of bytes available
    unsigned long   flClear = 0;    //  0 = leave stats, 1= clear
    char            achErrorMsg[100];
	char			*pszServerName = "elrayox";

    // check to see if the Open Server client has sent us an
    // attention.  If so, don't sent any results.

    if ( SRV_GOT_ATTENTION(pSrvProc) )
        return (0);
	

   
    // describe the result rows
    srv_describe ( pSrvProc,
                  1,
                  "server_name",
                  SRV_NULLTERM,
                  SRVCHAR,
                  CNLEN +1,
                  SRVCHAR,
                  0,
                  NULL);
   srv_describe ( pSrvProc,
                  2,
                  "timestamp",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
	 				   NULL);

    srv_describe ( pSrvProc,
                  3,
                  "sts0_start",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
   srv_describe ( pSrvProc,
                  4,
                  "sts0_jobsqueued",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
 
    srv_describe ( pSrvProc,
                  5,
                  "sts0_stimedout",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
  
    srv_describe ( pSrvProc,
                  6,
                  "sts0_pwerrors",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  

    // Send back result rows
    DosSemRequest( &StatsSem, -1L );

    for ( iCount = 0, i = giCurrentRow; 
	 						iCount<STAT_ARRAY_SIZE; iCount++, i--)
    {
	 	if (i < 0)
            i = STAT_ARRAY_SIZE - 1;
		if (aLanStats[i].Timestamp == 0)
      	break;    
        srv_setcoldata ( pSrvProc,
                  1,
                 	aLanStats[i].szServerName);
		  srv_setcollen (pSrvProc,
		  				1,
						strlen (aLanStats[i].szServerName));

          
        srv_setcoldata ( pSrvProc,
                  2,
                  &(aLanStats[i].Timestamp));

          
        srv_setcoldata ( pSrvProc,
                  3,
                  &(aLanStats[i].StatServ0.sts0_start));
        srv_setcoldata ( pSrvProc,
                  4,
                  &(aLanStats[i].StatServ0.sts0_jobsqueued));

  
  
        srv_setcoldata ( pSrvProc,
                  5,
                  &(aLanStats[i].StatServ0.sts0_stimedout));

  
        srv_setcoldata ( pSrvProc,
                  6,
                  &(aLanStats[i].StatServ0.sts0_pwerrors));

  

        
        // now send the row down the pipe
        if ( srv_sendrow( pSrvProc ) != SUCCEED )
        {
            // send back an error message and return
            srv_sendmsg( pSrvProc,
                     SRV_MSG_ERROR,
                     (DBINT) 1,
                     (DBTINYINT) 0,
                     (DBTINYINT) 0,
                     NULL,
                     0,
                     0,
                     "Error returning rows",
                     SRV_NULLTERM);
            srv_senddone (pSrvProc,
                      SRV_DONE_FINAL,
                      (DBUSMALLINT) 0,
                      (DBINT) 0);
            return ( SUCCEED );
        }

    }       
  	DosSemClear( &StatsSem );

    srv_senddone( pSrvProc,
               SRV_DONE_FINAL,
               (DBUSMALLINT) 0,
               (DBINT) 0);

    return (SUCCEED);
}

int	GetStatistics(pSrvProc)
SRV_PROC    FAR *pSrvProc;
{	
    char            *pszService = "SERVER";
    char            *pbBuffer;      // pointer to data buffer
    API_RET_TYPE    uStatus;
    int             i, iCount;         // Index for arguments
    unsigned short  cbBuffer;       // size of buffer
    unsigned short  cbTotalAvail;   // count of bytes available
    unsigned long   flClear = 0;    //  0 = leave stats, 1= clear
    char            achErrorMsg[100];
	char			*pszServerName = "elrayox";


    // check to see if the Open Server client has sent us an
    // attention.  If so, don't sent any results.

    if ( SRV_GOT_ATTENTION(pSrvProc) )
        return (0);
	

   
    // describe the result rows
    srv_describe ( pSrvProc,
                  1,
                  "server_name",
                  SRV_NULLTERM,
                  SRVCHAR,
                  CNLEN +1,
                  SRVCHAR,
                  strlen(pszServerName),
                  pszServerName);
   srv_describe ( pSrvProc,
                  2,
                  "timestamp",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
	 				   NULL);

    srv_describe ( pSrvProc,
                  3,
                  "sts0_start",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
    srv_describe ( pSrvProc,
                  4,
                  "sts0_fopens",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  5,
                  "sts0_devopens",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  6,
                  "sts0_jobsqueued",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  7,
                  "sts0_sopens",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  8,
                  "sts0_stimedout",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  9,
                  "sts0_serrorout",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  10,
                  "sts0_pwerrors",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  11,
                  "sts0_permerrors",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  12,
                  "sts0_syserrors",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  13,
                  "sts0_bytessent_low",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  14,
                  "sts0_bytessent_high",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  15,
                  "sts0_bytesrcvd_low",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  16,
                  "sts0_bytesrcvd_high",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
    srv_describe ( pSrvProc,
                  17,
                  "sts0_avresponse",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  18,
                  "sts0_reqbufneed",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
  
    srv_describe ( pSrvProc,
                  19,
                  "sts0_bigbufneed",
                  SRV_NULLTERM,
                  SRVINT4,
                  0,
                  SRVINT4,
                  4,
                  NULL);
    


    // Send back result rows
    DosSemRequest( &StatsSem, -1L );

    for ( iCount = 0, i = giCurrentRow; 
	 						iCount<STAT_ARRAY_SIZE; iCount++, i--)
    {
	 	  if (i < 0)
            i = STAT_ARRAY_SIZE - 1;
		if (aLanStats[i].Timestamp == 0)
      	break;    
        srv_setcoldata ( pSrvProc,
                  1,
                 	aLanStats[i].szServerName);
		  srv_setcollen (pSrvProc,
		  				1,
						strlen (aLanStats[i].szServerName));

          
        srv_setcoldata ( pSrvProc,
                  2,
                  &(aLanStats[i].Timestamp));

          
        srv_setcoldata ( pSrvProc,
                  3,
                  &(aLanStats[i].StatServ0.sts0_start));

        srv_setcoldata ( pSrvProc,
                  4,
                  &(aLanStats[i].StatServ0.sts0_fopens));

  
        srv_setcoldata ( pSrvProc,
                  5,
                  &(aLanStats[i].StatServ0.sts0_devopens));

  
        srv_setcoldata ( pSrvProc,
                  6,
                  &(aLanStats[i].StatServ0.sts0_jobsqueued));

  
        srv_setcoldata ( pSrvProc,
                  7,
                  &(aLanStats[i].StatServ0.sts0_sopens));

  
        srv_setcoldata ( pSrvProc,
                  8,
                  &(aLanStats[i].StatServ0.sts0_stimedout));

  
        srv_setcoldata ( pSrvProc,
                  9,
                  &(aLanStats[i].StatServ0.sts0_serrorout));

  
        srv_setcoldata ( pSrvProc,
                  10,
                  &(aLanStats[i].StatServ0.sts0_pwerrors));

  
        srv_setcoldata ( pSrvProc,
                  11,
                  &(aLanStats[i].StatServ0.sts0_permerrors));

  
        srv_setcoldata ( pSrvProc,
                  12,
                  &(aLanStats[i].StatServ0.sts0_syserrors));

  
        srv_setcoldata ( pSrvProc,
                  13,
                  &(aLanStats[i].StatServ0.sts0_bytessent_low));

  
        srv_setcoldata ( pSrvProc,
                  14,
                  &(aLanStats[i].StatServ0.sts0_bytessent_high));

    
        srv_setcoldata ( pSrvProc,
                  15,
                  &(aLanStats[i].StatServ0.sts0_bytesrcvd_low));

  
        srv_setcoldata ( pSrvProc,
                  16,
                  &(aLanStats[i].StatServ0.sts0_bytesrcvd_high));

        srv_setcoldata ( pSrvProc,
                  17,
                  &(aLanStats[i].StatServ0.sts0_avresponse));

  
        srv_setcoldata ( pSrvProc,
                  18,
                  &(aLanStats[i].StatServ0.sts0_reqbufneed));

  
        srv_setcoldata ( pSrvProc,
                  19,
                  &(aLanStats[i].StatServ0.sts0_bigbufneed));
        
        // now send the row down the pipe
        if ( srv_sendrow( pSrvProc ) != SUCCEED )
        {
            // send back an error message and return
            srv_sendmsg( pSrvProc,
                     SRV_MSG_ERROR,
                     (DBINT) 1,
                     (DBTINYINT) 0,
                     (DBTINYINT) 0,
                     NULL,
                     0,
                     0,
                     "Error returning rows",
                     SRV_NULLTERM);
            srv_senddone (pSrvProc,
                      SRV_DONE_FINAL,
                      (DBUSMALLINT) 0,
                      (DBINT) 0);
            return ( SUCCEED );
        }

    }       
  	DosSemClear( &StatsSem );

    srv_senddone( pSrvProc,
               SRV_DONE_FINAL,
               (DBUSMALLINT) 0,
               (DBINT) 0);

    return (SUCCEED);
}





void    StatDaemon()
{
    char            *pszService = "SERVER";
	 char					achServerName[100];
	 char            *pbBuffer;      // pointer to data buffer
    API_RET_TYPE    uStatus;
    int             iCount;         // Index for arguments
    unsigned short  cbBuffer;       // size of buffer
    unsigned short  cbTotalAvail;   // count of bytes available
    unsigned long   flClear = 0;    //  0 = leave stats, 1= clear

    while(1)
    {
        DosSemRequest( &StatsSem, -1L );
        if (giCurrentRow == STAT_ARRAY_SIZE - 1)
            giCurrentRow = 0;
        else
            giCurrentRow++;
            
    
        uStatus = NetStatisticsGet2("",
                                    pszService,
                                    0L,
                                    0,
                                    flClear,
                                    (char *) &aLanStats[giCurrentRow],
                                    sizeof(aLanStats[giCurrentRow]),
                                    &cbTotalAvail);

			if (uStatus != 0)
    			printf("error getting statistics.  Error number = %d\n",uStatus);

        strcpy(aLanStats[giCurrentRow].szServerName, "elrayox");
        time(&aLanStats[giCurrentRow].Timestamp);
     
    
    	DosSemClear( &StatsSem );
		DosSleep(gulSleepTime);
    }
    
}    


/*
** Name:  GetCustomerInfo)
**
** Purpose:  
**	This function implements a local stored procedure which retrieves
**  data from a simple data file.
** Input Parameters:
**	SRV_PROC	*pSrvProc; Pointer to server thread control structure
**
** Side Effects:
**	Sends messages and/or a result set to GDK server application client.
**
*/
                                 
RETCODE	GetCustomerInfo(pSrvProc)
SRV_PROC	FAR	*pSrvProc;
{	
	char 	achBuffer[BUF_SIZE];
	FILE	*pDataFile;
	char	*pchCustomerName, *pchCustomerAddress, *pchSales;


   
    // describe the columns in the result set
    srv_describe ( pSrvProc,   		// thread id for client
                  1,			  	// number of this column in result set
                  "Customer_Name",	// column name
                  SRV_NULLTERM,	 	// length of column 
                  SRVCHAR,			// datatype of column
                  MAX_NAME_LENGTH,	// maximum length of column
                  SRVCHAR,			// datatype of source 
                  0,				// length of source (0 if unknown at this time)
                  NULL);	// pointer to data

	 srv_describe ( pSrvProc,
                  2,
                  "Customer_Address",
                  SRV_NULLTERM,
					SRVCHAR,
                  MAX_ADDR_LENGTH,
                  SRVCHAR,
                  0,
                  NULL);

	 srv_describe ( pSrvProc,
                  3,
                  "FY90_SALES",
                  SRV_NULLTERM,
				  SRVCHAR,
                  16,
                  SRVCHAR,
                  0,
                  NULL);

	// if we were updating the file, we would control access using an
	// OS/2 semaphore, or use the file locking mechanisms supplied by
	// a PC ISAM product.

	pDataFile = fopen( pchFileName, "r"); 

	if (pDataFile == NULL)
	{
       // send back an error message and return
            srv_sendmsg( pSrvProc,
                     SRV_MSG_ERROR,
                     (DBINT) 1,
                     (DBTINYINT) 0,
                     (DBTINYINT) 0,
                     NULL,
                     0,
                     0,
                     "Couldn't open Customer data file",
                     SRV_NULLTERM);
          
	    srv_senddone( pSrvProc,
    	           SRV_DONE_FINAL,
        	       (DBUSMALLINT) 0,
            	   (DBINT) 0);

    	return (SUCCEED);
	}

	while (fgets(achBuffer, BUF_SIZE, pDataFile) != NULL)
	{
		printf("%s\n", achBuffer);
		// get the customer name
		// the data file uses a semicolon as a delimiter
		pchCustomerName = strtok(achBuffer, ";");
 	  	
		srv_setcoldata (pSrvProc, 1, pchCustomerName);
		// set the length of this column for this row
		srv_setcollen (pSrvProc,
		  				1,
						strlen (pchCustomerName));

		pchCustomerAddress = strtok(NULL, ";");
		
		srv_setcoldata (pSrvProc, 2, pchCustomerAddress);
		// set the length of this column for this row
		srv_setcollen (pSrvProc,
		  				2,
						strlen (pchCustomerAddress));
		
		// get the sales amount for this customer
		pchSales = strtok(NULL, ";");
		srv_setcoldata (pSrvProc, 3, pchSales);
		// set the length of this column for this row
		srv_setcollen (pSrvProc,
		  				3,
						strlen (pchSales));
	
 
    	    
        // now send the row down the pipe
        if ( srv_sendrow( pSrvProc ) != SUCCEED )
        {
            // send back an error message and return
            srv_sendmsg( pSrvProc,
                     SRV_MSG_ERROR,
                     (DBINT) 1,
                     (DBTINYINT) 0,
                     (DBTINYINT) 0,
                     NULL,
                     0,
                     0,
                     "Error returning Customer data",
                     SRV_NULLTERM);
			break;
        }

    }       
    srv_senddone( pSrvProc,
               SRV_DONE_FINAL,
               (DBUSMALLINT) 0,
               (DBINT) 0);
	fclose(pDataFile);
    return (SUCCEED);
}











#pragma check_stack()   // set stack checking to its default setting
